# -*- python -*-
# $RCSfile: displaymethods.py,v $
# $Revision: 1.121 $
# $Author: reida $
# $Date: 2007-05-15 20:51:17 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Simple display methods for Skeletons and Meshes.  Simpler than
# contour plots, in any case.

import sys

from oof2.SWIG.common import config
from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.common import threadstate
from oof2.SWIG.engine import mastercoord
from oof2.SWIG.engine import smallmatrix
from oof2.common import color
from oof2.common import debug
from oof2.common import mainthread
from oof2.common import parallel_enable
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common import utils
from oof2.common.IO import automatic
from oof2.common.IO import colormap
from oof2.common.IO import display
from oof2.common.IO import ghostgfxwindow
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import output
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump
from oof2.engine import mesh
from oof2.engine import skeletoncontext
from oof2.engine import skeletonmodifier
from oof2.engine.IO import outputDefs
import oof2.SWIG.engine.material

if parallel_enable.enabled():
    from oof2.SWIG.common import mpitools

FloatRangeParameter = parameter.FloatRangeParameter
IntRangeParameter = parameter.IntRangeParameter
AutoNumericParameter = parameter.AutoNumericParameter

##################

# Add a parameter to the graphics window's Settings menu, to control
# whether or not Mesh Elements with no Material are displayed.

def toggleEmptyDrawing(menuitem, hideEmpty):
    gfxwindow = menuitem.data
    gfxwindow.settings.hideEmptyElements = hideEmpty
    gfxwindow.draw()

def addToGfxSettings(gfxwindow):
    item = gfxwindow.menu.Settings.addItem(oofmenu.CheckOOFMenuItem(
        'Hide_Empty_Mesh_Elements',
        callback=toggleEmptyDrawing,
        value=gfxwindow.settings.hideEmptyElements,
        help="Toggle the display of elements with no Material.",
        discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/hideempty.xml')))
    item.data = gfxwindow

switchboard.requestCallback('open graphics window', addToGfxSettings)
ghostgfxwindow.defineGfxSetting('hideEmptyElements', True)

##################

# Skeleton and Mesh display methods are very similar, except for how
# they get some of their data.  These two base classes encapsulate the
# *differences* between Skeletons and Meshes, as far as displaying is
# concerned.  They're mixed in with other base classes to create the
# actual DisplayMethods.

class SkeletonDisplayMethod(display.DisplayMethod):
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def polygons(self, gfxwindow, skeleton):
        if parallel_enable.enabled():
            nodes = skeleton.all_skeletons["nodes"]
            elements = skeleton.all_skeletons["elements"]
            polys = []
            for i in range(mpitools.Size()):
                for el in elements[i]:
                    polys.append([primitives.Point(*nodes[i][ni])
                                  for ni in el])
            return  polys
        else:
            return [el.perimeter() for el in skeleton.element_iterator()]


# Dummy exception class, raised by
# _undisplaced_from_displaced_with_element if it overruns its
# iteration max.  This is expected to indicate failure due to range
# problems.  Caught and handled in the undisplaced_from_displaced
# funciton, which is the raising function's only caller.
class IterationMaxExceeded:
    pass
    
class MeshDisplayMethod(display.DisplayMethod):
    # A display method that displays data from a Mesh at positions
    # determined by the mesh.  self.where is a PositionOutput.
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def incomputable(self, gfxwindow):
        themesh = self.who().resolve(gfxwindow)
        return display.DisplayMethod.incomputable(self, gfxwindow) \
               or self.where.incomputable(themesh)
    def polygons(self, gfxwindow, themesh):


        # PARALLEL_RCL: Make changes here to display parallel mesh
        # There is an issue with clicking on the skeleton or mesh graphics:
        #  only the nodes or elements for the front-end process get the cursor or mark
        
        if parallel_enable.enabled():
            # This snippet taken from SkeletonDisplayMethod
            nodes = themesh.all_meshskeletons["nodes"]
            elements = themesh.all_meshskeletons["elements"]
            polys = []
            for i in range(mpitools.Size()):
                for el in elements[i]:
                    polys.append([primitives.Point(*nodes[i][ni])
                                  for ni in el])
            return polys
        else:
            if gfxwindow.settings.hideEmptyElements:
                edges = [element.perimeter()
                         for element in themesh.element_iterator()
                         if element.material() is not None]
            else:
                edges = [element.perimeter()
                         for element in themesh.element_iterator()]
            flatedges = utils.flatten(edges)
            # corners tells where on each edge to evaluate self.where
            corners = [[0.0]]*len(flatedges)
            # evaluate position output for all edges at once
            polys = self.where.evaluate(themesh, flatedges, corners)
            # give the corner list the same structure as the edge list: a
            # list of lists, where each sublist is the list of corners of
            # an element.
            polys = utils.unflatten(edges, polys)
            if len(polys) == 0:
                debug.fmsg(themesh)
                mainthread.runBlock(
                    reporter.warn,
                    ("No mesh elements drawn! Are there no materials assigned?",))
            return polys

    # Routines for converting between displaced and undisplaced
    # coordinates using this layer's PositionOutput object
    # ("self.where").  This is here primarily so that the mesh info
    # displays can be drawn in displaced coordinates -- since this
    # method refers to the topmost mesh display, only that display
    # (i.e. this object) knows the right PositionOutput to use.
    def displaced_from_undisplaced(self, gfxwindow, orig):
        meshctxt = self.who().resolve(gfxwindow)
        femesh = meshctxt.getObject()
        felem = meshctxt.enclosingElement(orig)
        return self._displaced_from_undisplaced_with_element(
            femesh, felem, orig)

    def _displaced_from_undisplaced_with_element(self, mesh, elem, orig):
        masterpos = elem.to_master(orig.x, orig.y)
        res = self.where.evaluate(mesh, [elem], [[masterpos]])
        return res[0]

    
    def _undisplaced_from_displaced_with_element(self, mesh, elem, pos):
        # Search master space for an x such that f(x)=pos, and return x.

        mtx = smallmatrix.SmallMatrix(2,2)
        rhs = smallmatrix.SmallMatrix(2,1)

        delta = 0.001 # Small compared to master space.
        delta_x = mastercoord.MasterCoord(delta, 0.0)
        delta_y = mastercoord.MasterCoord(0.0, delta)
        
        res = elem.center()

        done = False
        # Magic numbers.  The function can actually fail, because for
        # higher-order elements, the displacement mapping is
        # parabolic, and the range of the parabolic mapping might not
        # include the point pos, if it's too far away from the
        # element.  So, if we've gone for more than maxiters
        # iterations, raise an exception.
        tolerance = 1.0e-10
        maxiters = 50

        count = 0
        while not done:
            count += 1
            if count > maxiters:
                raise IterationMaxExceeded()

            fwd = self.where.evaluate(mesh, [elem],[[res]])[0]
            fwddx = self.where.evaluate(mesh, [elem],[[res+delta_x]])[0]
            fwddy = self.where.evaluate(mesh, [elem],[[res+delta_y]])[0]
            
            dfdx = (fwddx-fwd)/delta
            dfdy = (fwddy-fwd)/delta
            diff = pos-fwd

            rhs.setitem(0,0,diff[0])
            rhs.setitem(1,0,diff[1])
            
            mtx.setitem(0,0,dfdx[0])
            mtx.setitem(0,1,dfdy[0])
            mtx.setitem(1,0,dfdx[1])
            mtx.setitem(1,1,dfdy[1])

            ## TODO: For a 2x2 matrix, is it faster to write out the
            ## solution, rather than using a general purpose routine?
            r = mtx.solve(rhs)

            resid = (rhs[0,0]**2+rhs[1,0]**2)
            res = mastercoord.MasterCoord(res[0]+rhs[0,0],
                                          res[1]+rhs[1,0])

            if resid<tolerance:
                done = True

        return res

    # TODO LATER: This whole process leans fairly hard on the
    # single-point, single-element version of self.where.evaluate() --
    # performance improvements there would be welcome.
    def undisplaced_from_displaced(self, gfxwindow, pos):
        meshctxt = self.who().resolve(gfxwindow)
        femesh = meshctxt.getObject()

        # Searches over the list of elements, sorted in order from
        # nearest to farthest (as measured by the center points) from
        # the click position.  This is much less expensive than
        # running the Newton-Raphson iteration extra useless times.
        # TODO LATER: This is slow for large meshes, and in
        # particular, is very slow to fail.  Find a cleverer way to
        # select which elements to search.
        ellist = []
        ei = femesh.element_iterator()
        while not ei.end():
            el = ei.element()
            if el.material() is not None:
                distance2 = (self.where.evaluate(
                    femesh, [el],[[el.center()]])[0] - pos)**2
                ellist.append( (distance2, el) )
            ei.next()
        ellist.sort()

        smallestdist = None
        smallestres = None
        smallestel = None
        for dist2, el in ellist:
            if el.material() is not None:
                try:
                    res = self._undisplaced_from_displaced_with_element(
                        femesh, el, pos)
                except IterationMaxExceeded:
                    # Newton-Raphson failed, probable range error, go
                    # on to the next element.
                    pass
                else:
                    # We got a result.  Find how far it is outside the
                    # master element
                    dist = el.masterelement().outOfBounds(res)

                    # If it's inside, then we're done
                    if dist <= 0.0:
                        return el.from_master(res[0], res[1])
                    # If it's outside, it may just be due to round off.
                    # We have to wait until we've looked at all the
                    # elements to make sure, though.
                    if smallestdist is None or dist < smallestdist:
                        smallestdist = dist
                        smallestres = res
                        smallestel = el

        # If we got here, then the point isn't inside any element.
        if smallestdist < 1.e-10:
            if smallestel:
                # It's close enough, though...
                return smallestel.from_master(smallestres[0], smallestres[1])

        raise ooferror.ErrBoundsError("No element found")
        
###########################

## Default values of display parameters for SkeletonEdgeDisplay and
## MeshEdgeDisplay, and menu items to set them.

defaultSkeletonWidth = 0
defaultSkeletonColor = color.black
defaultMeshWidth = 0
defaultMeshColor = color.black

def _setDefaultSkeletonEdgeParams(menuitem, color, width):
    global defaultSkeletonColor
    defaultSkeletonColor = color
    global defaultSkeletonWidth
    defaultSkeletonWidth = width
    
mainmenu.gfxdefaultsmenu.Skeletons.addItem(oofmenu.OOFMenuItem(
    'Skeleton_Edges',
    callback=_setDefaultSkeletonEdgeParams,
    ordering = 0,
    params=[color.ColorParameter('color', defaultSkeletonColor,
                                 tip=parameter.emptyTipString),
            IntRangeParameter('width', (0,10), defaultSkeletonWidth,
                              tip="Line thickness, in pixels.")
            ],
    help="Set the default parameters for Skeleton edge displays.",
    discussion="""<para>

    Set the default parameters for
    <link linkend="RegisteredClass:SkeletonEdgeDisplay"><classname>SkeletonEdgeDisplays</classname></link>.
    See <xref linkend="RegisteredClass:SkeletonEdgeDisplay"/> for the details.      This command may be placed in the &oof2rc; file
    to set a default value for all &oof2; sessions.

    </para>"""))

def _setDefaultMeshEdgeParams(menuitem, color, width):
    global defaultMeshColor
    defaultMeshColor = color
    global defaultMeshWidth
    defaultMeshWidth = width

mainmenu.gfxdefaultsmenu.Meshes.addItem(oofmenu.OOFMenuItem(
    'Mesh_Edges',
    callback=_setDefaultMeshEdgeParams,
    ordering=0,
    params=[color.ColorParameter('color', defaultMeshColor,
                                 tip=parameter.emptyTipString),
            IntRangeParameter('width', (0,10), defaultMeshWidth,
                              tip="Line thickness, in pixels.")],
    help="Set the default parameters for Mesh edge displays.",
    discussion="""<para>

    Set the default parameters for
    <link linkend="RegisteredClass:MeshEdgeDisplay"><classname>MeshEdgeDisplays</classname></link>.
    See <xref linkend="RegisteredClass:MeshEdgeDisplay"/> for the details.
    This command may be placed in the &oof2rc; file to set a default value
    for all &oof2; sessions.

    </para>"""))

####

class EdgeDisplay:
    def draw(self, gfxwindow, device):
        themesh = self.who().getObject(gfxwindow)
        polygons = self.polygons(gfxwindow, themesh)
        device.comment("EdgeDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        #import pdb
        #pdb.set_trace()
        for polygonset in polygons:
            device.draw_polygon(primitives.Polygon(polygonset))

class MeshEdgeDisplay(EdgeDisplay, MeshDisplayMethod):
    # EdgeDisplay draws the edges of the Elements
    def __init__(self, where, width=defaultMeshWidth, color=defaultMeshColor):
        if where is not None:
            self.where = where.clone()
        else:
            self.where = None           # ?
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        EdgeDisplay.draw(self, gfxwindow, device)
    def getTimeStamp(self, gfxwindow):
        return max(
            MeshDisplayMethod.getTimeStamp(self, gfxwindow),
            gfxwindow.settings.getTimeStamp('hideEmptyElements'))

class SkeletonEdgeDisplay(EdgeDisplay, SkeletonDisplayMethod):
    def __init__(self, width=defaultSkeletonWidth, color=defaultSkeletonColor):
        self.width = width
        self.color = color
        SkeletonDisplayMethod.__init__(self)

registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             MeshEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                            tip="Plot at displaced or original position?"),
    color.ColorParameter('color',
                         defaultMeshColor,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), defaultMeshWidth,
                      tip="Line thickness, in pixels.")],
                             whoclasses = ('Mesh',),
                             tip="Draw the edges of Mesh Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/meshedgedisplay.xml')
                             )
    
registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             SkeletonEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    color.ColorParameter('color',
                         defaultSkeletonColor,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), defaultSkeletonWidth,
                      tip="Line thickness, in pixels.")],
                             whoclasses = ('Skeleton',),
                             tip="Draw the edges of Skeleton Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonedgedisplay.xml')
                             )

######################

class PerimeterDisplay(MeshDisplayMethod):
    def __init__(self, where, width=0, color=color.black):
        self.where = where.clone()
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        themesh = self.who().resolve(gfxwindow)
        femesh = themesh.getObject()
        device.comment("PerimeterDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        for element in femesh.element_iterator():
            el_edges = element.perimeter()
            for edge in el_edges:
                if element.exterior(edge.startpt(), edge.endpt()):
                    pts = self.where.evaluate(femesh, [edge], [[0.0, 1.0]])
                    device.draw_segment(primitives.Segment(pts[0], pts[1]))

if config.devel()>=1:
    # This doesn't work as expected...
    registeredclass.Registration('Perimeter',
                                 display.DisplayMethod,
                                 PerimeterDisplay,
                                 ordering=1.0,
                                 layerordering=display.SemiLinear,
                                 params=[
        output.PositionOutputParameter('where', outputDefs.originalPosition,
                                tip="Plot at displaced or original position?"),
        color.ColorParameter('color', color.black,
                             tip=parameter.emptyTipString),
        IntRangeParameter('width', (0,10), 0,
                          tip="Line width.")
        ],
                                 whoclasses = ('Mesh',),
                                 tip="Outline the perimeter of the Mesh",
                                 discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/perimeterdisplay.xml')
                 )

######################

##class CenterFillDisplay:
##    # CenterFillDisplay fills each Element with a solid color,
##    # determined by the 'what' at the center of the Element.

##    def incomputable(self, gfxwindow):
##        themesh = self.who().resolve(gfxwindow)
##        return self.what.incomputable(themesh) or \
##               MeshDisplayMethod.incomputable(self, gfxwindow)
##    def draw(self, gfxwindow, device):
##        themesh = self.who().resolve(gfxwindow).mesh()
##        device.comment("SolidFill")
##        device.set_colormap(self.colormap)
##        polygons = self.polygons(themesh)
##        elements = tuple(themesh.element_iterator())
##        evaluationpoints = [[el.center()] for el in elements]
##        values = self.what.evaluate(themesh, elements, evaluationpoints)
##        max_value = max(values)
##        min_value = min(values)
##        # HACK: If the values are equal, put the actual value in
##        # the middle of an arbitrary range.
##        if max_value==min_value:
##            max_value += 1.0
##            min_value -= 1.0
##        for polygon, value in zip(polygons, values):
##            if value is not None:
##                device.set_fillColor((value-min_value)/(max_value-min_value))
##                device.fill_polygon(primitives.Polygon(polygon))
        
##class MeshCenterFillDisplay(CenterFillDisplay, MeshDisplayMethod): 
##    def __init__(self, where, what, colormap=colormap.ThermalMap()):
##        self.where = where.clone()
##        self.what = what.clone()
##        self.colormap = colormap
##        MeshDisplayMethod.__init__(self)


##registeredclass.Registration('Solid Fill',
##                             display.DisplayMethod,
##                             MeshCenterFillDisplay,
##                             ordering=2.0,
##                             layerordering=display.Planar,
##                             params=[
##    output.ScalarOutputParameter('what', tip='Quantity to be plotted'),
##    output.PositionOutputParameter('where', outputDefs.originalPosition,
##                            tip='Plot at original or displaced position?'),
##    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap(),
##              tip="Fill color.")],
##                             whoclasses = ('Mesh',),
##                             tip="Quick and dirty contour plot.",
##                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/centerfilldisplay.xml')
##                             )

# SkeletonCenterFillDisplay isn't useful unless there are Outputs that
# can be evaluated on a Skeleton.

##class SkeletonCenterFillDisplay(CenterFillDisplay, SkeletonDisplayMethod): 
##    def __init__(self, what, colormap=colormap.ThermalMap()):
##        self.what = what
##        self.colormap = colormap

##registeredclass.Registration('SolidFill',
##             display.DisplayMethod,
##             SkeletonCenterFillDisplay,
##             ordering=2.01,
##             params=[
##    ScalarOutputParameter('what', tip='Quantity to be plotted'),
##    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap())],
##             whoclasses = ('Mesh')
##             )

######################

class MaterialDisplay:
    def draw(self, gfxwindow, device):
        device.comment("Material Color")
        themesh = self.who().resolve(gfxwindow).getObject()
        polygons = self.polygons(gfxwindow, themesh)
        # colorcache is a dictionary of colors keyed by Material.  It
        # prevents us from having to call material.fetchProperty for
        # each element.
        colorcache = {}
        for polygon, material in zip(polygons,
                                     self.materials(gfxwindow, themesh)):
            if material is not None:
                try:
                    # If material has been seen already, retrieve its color.
                    color = colorcache[material]
                except KeyError:
                    # This material hasn't been seen yet.
                    try:
                        colorprop = material.fetchProperty('Color')
                        color = colorprop.color()
                    except ooferror.ErrNoSuchProperty:
                        color = None
                    colorcache[material] = color
                if color is not None:
                    device.set_fillColor(color)
                    device.fill_polygon(primitives.Polygon(polygon))
    def getTimeStamp(self, gfxwindow):
        microstructure = self.who().resolve(gfxwindow).getMicrostructure()
        return max(display.DisplayMethod.getTimeStamp(self, gfxwindow),
                   oof2.SWIG.engine.material.getMaterialTimeStamp(microstructure))
                    
class SkeletonMaterialDisplay(MaterialDisplay, SkeletonDisplayMethod):
    def __init__(self):
        SkeletonDisplayMethod.__init__(self)
    def materials(self, gfxwindow, themesh):
        return [element.material(themesh)
                for element in themesh.element_iterator()]

class MeshMaterialDisplay(MaterialDisplay, MeshDisplayMethod):
    def __init__(self, where):
        self.where = where.clone()
        MeshDisplayMethod.__init__(self)
    def materials(self, gfxwindow, themesh):
        # Because MeshDisplayMethod.polygons only returns the borders
        # of elements with an assigned material, this should only
        # return the non-trivial materials.
        allmats = [element.material()
                   for element in themesh.element_iterator()]
        if gfxwindow.settings.hideEmptyElements:
            return filter(None, allmats)
        return allmats
    def getTimeStamp(self, gfxwindow):
        return max(MaterialDisplay.getTimeStamp(self, gfxwindow),
                   gfxwindow.settings.getTimeStamp('hideEmptyElements'))
    
    
registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             SkeletonMaterialDisplay,
                             layerordering=display.Planar(1),
                             ordering=0.1,
                             params=[],
                             whoclasses=('Skeleton',),
                             tip="Fill each Element with the color of its assigned Material.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonmaterialdisplay.xml')
                             )

registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             MeshMaterialDisplay,
                             ordering=0.11,
                             layerordering=display.Planar(2),
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                           tip='Plot at original or displaced position?')],
                             whoclasses=('Mesh',),
                             tip="Fill each Element with the color of its assigned Material.",
                                                          discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/meshmaterialdisplay.xml')
                             )

###########################################

class SkeletonQualityDisplay(SkeletonDisplayMethod):
    contourmaplevels = 32
    def __init__(self, alpha, min, max, colormap):
        self.alpha = alpha
        self.colormap = colormap
        self.max = max
        self.min = min
        self.vmax = None
        self.vmin = None
        self.contourmaphidden = 0
        self.lock = lock.Lock()
        SkeletonDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        self.lock.acquire()
        try:
            device.comment("Skeleton Energy")
            skel = self.who().resolve(gfxwindow).getObject()
            # get polygons and element energy in one pass
            polyenergy = [(el.perimeter(), el.energyTotal(skel, self.alpha))
                        for el in skel.element_iterator()
                        if not el.illegal()]
            # find actual range of data
            self.vmax = self.vmin = polyenergy[0][1]
            for (p,e) in polyenergy[1:]:
                if e > self.vmax:
                    self.vmax = e
                if e < self.vmin:
                    self.vmin = e
            # Set plot limits to either the actual data extremes, or
            # to the passed in values.  Store the actual limits in
            # vmin and vmax.
            if self.max == automatic.automatic:
                max = self.vmax
            else:
                max = self.max
                self.vmax = max
            if self.min == automatic.automatic:
                min = self.vmin
            else:
                min = self.min
                self.vmin = min
            if max == min:
                max += 1.0
                self.vmax += 1.0
                min -= 1.0
                self.vmin -= 1.0
            device.set_colormap(self.colormap)
            for polygon, energy in polyenergy:
                device.set_fillColor((energy-min)/(max-min))
                device.fill_polygon(primitives.Polygon(polygon))
        finally:
            self.lock.release()
    def getTimeStamp(self, gfxwindow):
        skelcontext = self.who().resolve(gfxwindow)
        return max(self.timestamp,
                   skelcontext.getTimeStamp(gfxwindow),
                   skelcontext.getMicrostructure().getTimeStamp())
    def contour_capable(self, gfxwindow):
        return not self.incomputable(gfxwindow)
    def get_contourmap_info(self):
        if self.vmax is not None:
            delta = (self.vmax - self.vmin)/(self.contourmaplevels-1.)
            return (self.vmin, self.vmax,
                    [self.vmin+x*delta for x in range(self.contourmaplevels)])
        return (0., 1., [0])
    def hide_contourmap(self):
        self.contourmaphidden = 1
    def show_contourmap(self):
        self.contourmaphidden = 0
    def draw_contourmap(self, gfxwindow, device):
        self.lock.acquire()
        try:
            if self.vmax is not None:
                aspect_ratio = gfxwindow.settings.aspectratio
                height = self.vmax - self.vmin
                width = height/aspect_ratio
                delta = height/(self.contourmaplevels-1.)
                device.comment("Colorbar minimum: %g" % self.vmin)
                device.comment("Colorbar maximum: %g" % self.vmax)
                device.set_colormap(self.colormap)
                for i in range(self.contourmaplevels):
                    low = i*delta
                    high = (i+1)*delta
                    rect_bdy = [primitives.Point(0.0, low),
                                primitives.Point(0.0, high),
                                primitives.Point(width, high),
                                primitives.Point(width, low)]
                    rectangle = primitives.Polygon(rect_bdy)
                    if height > 0.0:
                        device.set_fillColor(low/height)
                    else:
                        device.set_fillColor(0.0)
                    device.fill_polygon(rectangle)
        finally:
            self.lock.release()
    

registeredclass.Registration(
    'SkeletonQuality',
    display.DisplayMethod,
    SkeletonQualityDisplay,
    ordering=100,
    layerordering=display.Planar(1.1),
    whoclasses=('Skeleton',),
    params=[skeletonmodifier.alphaParameter,
            parameter.RegisteredParameter('colormap', colormap.ColorMap,
                                          colormap.ThermalMap(),
                                          tip="color scheme"),
            AutoNumericParameter('min', automatic.automatic,
                               tip="lowest energy to display, or 'automatic'"),
            AutoNumericParameter('max', automatic.automatic,
                              tip="highest energy to display, or 'automatic'")
            ],
    tip="Color each element according to its effective energy.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skelqualdisplay.xml')
    )

###########################################

def defaultSkeletonEdgeDisplay():
    return SkeletonEdgeDisplay(color=defaultSkeletonColor,
                               width=defaultSkeletonWidth)

ghostgfxwindow.DefaultLayer(skeletoncontext.skeletonContexts,
                            defaultSkeletonEdgeDisplay)

def defaultMeshEdgeDisplay():
    return MeshEdgeDisplay(where=outputDefs.actualPosition,
                           color=defaultMeshColor,
                           width=defaultMeshWidth)

ghostgfxwindow.DefaultLayer(mesh.meshes, defaultMeshEdgeDisplay)
