# -*- python -*-
# $RCSfile: boundary.py,v $
# $Revision: 1.64 $
# $Author: vrc $
# $Date: 2007-03-07 23:07:12 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# A note on EdgeBoundary, PointBoundary, and Profile classes.
#
# A Profile is an as-generic-as-possible way of getting values of some
# function out, given some indiciation of where to evaluate.  The
# indication is given in several forms at once -- the index of the
# current node (conventionaly denoted by i), the (x,y) location of the
# current evaluation, the distance (conventionally denoted s) along
# the current boundary, and the fractional distance (conventionally,
# alpha) along the current boundary.  The location information may
# include orientation information about the current normal direction
# of the boundary, (nx, ny).  (x,y) and "alpha" will always be
# provided.  "i" will be provided by PointBoundaries, and "s" and (nx,
# ny) will be provided by EdgeBoundaries.
#
# EdgeBoundaries are smooth sets of edges, and expect to be able
# to evaluate their corresponding conditions' profiles by passing
# in (x,y), (nx, ny), s, and alpha.
#

# PointBoundaries are sets of nodes, and expect to evaluate their
# profiles by passing the integer index i of the node within the
# boundary, along with its (x,y) location and "alpha".  The meaning of
# "alpha" in this context is that it is 0 for the first node, 1 for
# the last node, and progresses linearly with nodes, except for a
# PointBoundary with exactly one node, where it is 0.
#
# As currently instantiated, it's not possible to convert
# continuum boundaries into point boundaries, even though
# continuum boundaries have a nodal representation.  In
# principle, a flux boundary condition can be expressed in
# terms of equivalent forces at nodes, which we might want to
# do later.
#
# Both kinds of boundaries must provide invokeFixed and invokeFloat,
# but invokeFlux and invokeForce are optional.

from oof2.SWIG.common import ooferror
from oof2.SWIG.engine import edgeset
from oof2.common import debug
from oof2.common import registeredclass
from oof2.engine import profile
import copy
import oof2.engine.mesh
ErrSetupError = ooferror.ErrSetupError


# This class describes the value of a profile (associated with a
# condition) at a particular node.  It's primarily used for
# intersection checking, and allows the actual evaluation of the
# profile function (presumed to be expensive) to be deferred until a
# coincidence of location and field are certain.  The associated
# boundary may be either Edge or Point, this class doesn't care.


class LocatedCondition:
    def __init__(self, bc, here = None, offset = 0.0):
        self.condition = bc
        self.here = here # "Location" object, promises .position member.
        self.offset = offset
    def __call__(self):
        if self.here:
            return self.condition.profile_obj(self.here) + self.offset
        else:
            raise ErrSetupError("No location specified in LocatedCondition.")
        

#*=-=*##*=-=*##*=-=*##*=-=#*=-=*##*=-=*##*=-=*##*=-=*##*=-=*##*=-=*##*=-=*#

# Parent class of EdgeBoundary and PointBoundary, has a few simple
# global operations.  Fixed and float conditions are stored explicitly
# because they need to be checked for collisions.
class Boundary:
    def __init__(self, name, mesh, visible=True):
        self._name = name
        self.mesh = mesh # C Femesh instance.
        self.floatConditions = []
        self.fixedConditions = []
        self.allConditions = {}
        self.visible = visible
    #
    # Redefine "size" in subclasses.
    def size(self):
        pass
    
    def name(self):
        return self._name

    def rename(self, name):
        self._name = name

    # double-dispatch function for checking compatibility of the given
    # condition with existing conditions.
    def checkCondition(self, condition, exclude):
        return condition.checkBoundary(self, exclude)
        
    def addCondition(self, condition):
        self.allConditions[condition.name]=condition
        condition.addToBoundary(self)

    # Provide benign implementations of the optional invocations.
    def invokeFlux(self, subproblem, time):
        pass

    def invokeForce(self, subproblem, time):
        pass
    
    #
    # Reset function for stateful boundary conditions.  Removes
    # fixed BC's added by intersection with floats, reactivates
    # deactivated floats, and clears FloatBC internal dictionaries.
    def reset(self):
        for fixed in self.fixedConditions[:]:
            if fixed.promotion():
                fixed.remove_from_mesh()
        for float in self.floatConditions[:]:
            float.reset()

   
# An EdgeBoundary object has as its geometry an Edgeset.
#
# See comments at PointBoundary for info about maintaining consistency.
#
########################################################################
########################################################################

class EdgeBoundary(Boundary):
    def __init__(self, name, mesh, eset = None):
        Boundary.__init__(self, name, mesh)
        if eset == None:
            self.edgeset = edgeset.EdgeSet(self.mesh)
        else:
            self.edgeset = eset 
        self.fluxConditions = []
        self.periodicConditions = []
    def size(self):
        return self.edgeset.size()
    # addFloat and addFixed should maybe check for self-intersections.
    def __repr__(self):
        outstr = "\nEdgeBoundary("
        for float in self.floatConditions:
            outstr += float.__repr__()
        for fixed in self.fixedConditions:
            outstr += fixed.__repr__()
        for flux in self.fluxConditions:
            outstr += flux.__repr__()
        outstr += ")\n"
        return outstr
    #
    def clearConditions(self):
        self.floatConditions = []
        self.fixedConditions = []
        self.fluxConditions = []
        self.periodicConditions = []

    # This routine needs to live in the child class, because this
    # class knows what sorts of conditions we have.  Remove the
    # condition in whichever list it occurs.
    def removeCondition(self, condition):
        del self.allConditions[condition.name]
        # Removal doesn't require special processing.
        try:
            self.floatConditions.remove(condition)
        except ValueError:
            pass
        try:
            self.fixedConditions.remove(condition)
        except ValueError:
            pass
        try:
            self.fluxConditions.remove(condition)
        except ValueError:
            pass
        try:
            self.periodicConditions.remove(condition)
            # will need to remove all corresponding float conditions
        except ValueError:
            pass


    #
    # AddEdge, gateway to the EdgeSet.
    def addEdge(self, boundaryedge):
        self.edgeset.addEdge(boundaryedge)

    def addFloatCondition(self, condition):
        self.floatConditions.append(condition)
    def checkFloatCondition(self, condition, exclude):
        for old in self.floatConditions:
            if old is not exclude and old.intersect(condition) and not old.is_explicitly_disabled():
                return "Float BC collision on edge boundary"
        return 0

    def addPeriodicCondition(self, condition):
        self.periodicConditions.append(condition)
    def checkPeriodicCondition(self, condition, exclude):
        for old in self.periodicConditions:
            if old is not exclude and old.intersect(condition) and not old.is_explicitly_disabled():
                return "Periodic BC collision on edge boundary"
        return 0


    def whichPeriodicEdge(self, mesh_obj):
        # Helper function returns which periodic outer edge this
        # boundary spans and it's direction.  If the edge boundary in
        # question does not span a periodic outer edge, the function
        # returns 0.
        skel = mesh_obj.skeleton
        nodes = self.edgeset.nodes()
        if nodes[0].position().x == nodes[1].position().x and skel.left_right_periodicity:
            if nodes[0].position().y == 0 and nodes[-1].position().y == skel.MS.size()[1]:
                direction = 1
            elif nodes[-1].position().y == 0 and nodes[0].position().y == skel.MS.size()[1]:
                direction = -1
            else:
                return 0
            for i in xrange(2,len(nodes)):
                if not nodes[i].position().x == nodes[i-1].position().x:
                    return 0
            return ('x',nodes[0].position().x,direction)
        if nodes[0].position().y == nodes[1].position().y and skel.top_bottom_periodicity:
            if nodes[0].position().x == 0 and nodes[-1].position().x == skel.MS.size()[0]:
                direction = 1
            elif nodes[-1].position().x == 0 and nodes[0].position().x == skel.MS.size()[0]:
                direction = -1
            else:
                return 0
            for i in xrange(2,len(nodes)):
                if not nodes[i].position().y == nodes[i-1].position().y:
                    return 0
            return ('y',nodes[0].position().y,direction)


    def addFixedCondition(self, condition):
        self.fixedConditions.append(condition)
    def checkFixedCondition(self, condition, exclude):
        for old in self.fixedConditions:
            if old is not exclude and old.intersect(condition) and not old.is_explicitly_disabled():
                return "Dirichlet BC collision on edge boundary"
        return 0
        
    def addFluxCondition(self, condition):
        self.fluxConditions.append(condition)
    def checkFluxCondition(self, condition, exclude):
        return 0                        # ok

    def addForceCondition(self, condition):
        raise ErrSetupError("Edge boundaries do not support force conditions.")
    def checkForceCondition(self, condition, exclude):
        return "Edge boundaries do not support force conditions."
    #
    def invokeFixed(self, subproblem, time):
        for (node, location) in self.edgeset.locations():
            if subproblem.containsNode(node):
                location.set_time(time)
                for bc in self.fixedConditions:
                    bc(subproblem, node, bc.profile_obj(location))
    def invokeFloat(self, subproblem, time):
        for (node, location) in self.edgeset.locations():
            if subproblem.containsNode(node):
                location.set_time(time)
                for bc in self.floatConditions:
                    bc(subproblem, node, bc.profile_obj(location))
        for bc in self.floatConditions:
            bc.contrib_rhs(subproblem)
    # invoke periodic here?
    def expandFloat(self, subproblem, time):
        for (node, location) in self.edgeset.locations():
            if subproblem.containsNode(node):
                location.set_time(time)
                for bc in self.floatConditions:
                    # print "Expanding at a node."
                    bc.expand(subproblem, node, bc.profile_obj(location))
    def invokeFlux(self, subproblem, time):
        for bc in self.fluxConditions:
            bc(subproblem, self.edgeset, bc.profile_obj, time)

    # Functions that help out with intersection-detection.
    # Return a dictionary of nodes whose values are lists of
    # LocatedCondition objects.  A "LocatedCondition" knows how
    # to evaluate the condition's profile at this node.
    
    def getFixedGeometry(self, subproblem, time):
        redict = {}
        if self.fixedConditions:
            for (node, location) in self.edgeset.locations():
                location.set_time(time)
                reslist = []
                for fixedbc in self.fixedConditions:
                    # do not include disabled bc's
                    if not fixedbc.is_disabled(subproblem):
                        reslist.append(
                            LocatedCondition(fixedbc, location,
                                             offset = fixedbc.offset ))
                redict[node] = reslist
        return redict
    
    def getFloatGeometry(self, subproblem, time):
        redict = {}
        if self.floatConditions:
            for (node, location) in self.edgeset.locations():
                location.set_time(time)
                reslist = []
                for floatbc in self.floatConditions:
                    # do not include disabled bc's
                    if not floatbc.is_disabled(subproblem):
                        reslist.append(LocatedCondition(floatbc, location))
                redict[node] = reslist
        return redict
  

#
# PointBoundary
#
########################################################################
########################################################################

class PointBoundary(Boundary):
    def __init__(self, name, mesh, nodeset=None, visible=True):
        Boundary.__init__(self, name, mesh, visible)
        if nodeset:
            self.nodeset = nodeset
        else:
            self.nodeset = NodeSet(mesh)
        self.forceConditions = []
    # Generic "addNode" capability might be useful -- if you just want
    # to group nodes, e.g. corners, and then do the same thing to all
    # of them, it may not be necessary to distinguish between individual
    # nodes on the basis of keys.  For this case, auto-generate a key
    # and use it to insert into the dictionary.  
    #
    # Intersection/consistency checking is now done at assignment time
    # by the boundaries themselves.  To guarantee continued consistency,
    # users and/or the UI should always assign all fixed boundaries
    # prior to assigning any floating boundaries.
    def size(self):
        return len(self.nodeset)
    def __repr__(self):
        outstr = "PointBoundary("
        for float in self.floatConditions:
            outstr += float.__repr__()
        for fixed in self.fixedConditions:
            outstr += fixed.__repr__()
        for force in self.forceConditions:
            outstr += force.__repr__()
        outstr += ")"
        return outstr
    #
    def clearConditions(self):
        self.floatConditions = []
        self.fixedConditions = []
        self.forceConditions = []
    #
    def removeCondition(self, condition):
        del self.allConditions[condition.name]
        # Removal doesn't require special processing.
        try:
            self.floatConditions.remove(condition)
        except ValueError:
            pass
        try:
            self.fixedConditions.remove(condition)
        except ValueError:
            pass
        try:
            self.forceConditions.remove(condition)
        except ValueError:
            pass
    #
    def addNode(self, node):
        self.nodeset.addNode(node)

    def addFloatCondition(self, condition):
        self.floatConditions.append(condition)
    def checkFloatCondition(self, condition, exclude):
        for old in self.floatConditions:
            if old is not exclude and old.intersect(condition) and not old.is_explicitly_disabled():
                return "Floating BC collision on point boundary."
        return 0
    
    def addFixedCondition(self, condition):
        self.fixedConditions.append(condition)
    def checkFixedCondition(self, condition, exclude):
        for old in self.fixedConditions:
            if old is not exclude and old.intersect(condition) and not old.is_explicitly_disabled():
                return "Dirichlet BC collision on point boundary."
        return 0

    def addFluxCondition(self, condition):
        raise ErrSetupError("Point boundaries do not support flux BC's.")
    def checkFluxCondition(self, condition, exclude):
        return "Point boundaries do not support flux BC's."

    def addPeriodicCondition(self, condition):
        raise ErrSetupError("Point boundaries do not support periodic BC's.")
    def checkPeriodicCondition(self, condition, exclude):
        return "Point boundaries do not support periodic BC's."

    def addForceCondition(self, condition):
        self.forceConditions.append(condition)
    def checkForceCondition(self, condition, exclude):
        return 0                        # ok

    def invokeFixed(self, subproblem, time):
        for (node, location) in self.nodeset.locations():
            location.set_time(time)
            for bc in self.fixedConditions:
                bc(subproblem, node, bc.profile_obj(location))
    def invokeFloat(self, subproblem, time):
        for (node, location) in self.nodeset.locations():
            location.set_time(time)
            for bc in self.floatConditions:
                bc(subproblem, node, bc.profile_obj(location))
        for bc in self.floatConditions:
            bc.contrib_rhs(subproblem)
    def expandFloat(self, subproblem, time):
        for (node, location) in self.nodeset.locations():
            location.set_time(time)
            for bc in self.floatConditions:
                bc.expand(subproblem, node, bc.profile_obj(location))

    def invokeForce(self, subproblem, time):
        for (node, location) in self.nodeset.locations():
            location.set_time(time)
            for bc in self.forceConditions:
                bc(subproblem, node, bc.profile_obj(location))

    # Returns a dictionary of all the nodes on the boundary, with
    # an accompanying list of all the fixed boundary conditions,
    # with their unevaluated profile functions.
    def getFloatGeometry(self, subproblem, time):
        redict = {}
        if self.floatConditions:
            for (node, location) in self.nodeset.locations():
                location.set_time(time)
                reslist = []
                for floatbc in self.floatConditions:
                    # do not include disabled bc's
                    if not floatbc.is_disabled(subproblem):
                        reslist.append(LocatedCondition(floatbc, location ))
                redict[node] = reslist
        return redict
    def getFixedGeometry(self, subproblem, time):
        redict = {}
        if self.fixedConditions:
            for (node, location) in self.nodeset.locations():
                location.set_time(time)
                reslist = []
                for fixedbc in self.fixedConditions:
                    # do not include disabled bc's
                    if not fixedbc.is_disabled(subproblem):
                        reslist.append(
                            LocatedCondition(fixedbc, location,
                                             offset = fixedbc.offset))
                redict[node] = reslist
        return redict



# NodeSet class -- provides an ordered list of nodes, and knows their
# indices.  Can also provide a "location" service, which returns the
# appropriate object.  
class NodeSet:
    def __init__(self, mesh):
        self.mesh = mesh
        self.data = []
    # Since data is a list, key must be an integer.
    def __getitem__(self, key):
        return self.data[key]
    def __len__(self):
        return len(self.data)
    def addNode(self, node):
        self.data.append(node)
    def removeNode(self, node):
        self.data.remove(node)
        
    # Return a list of location objects, one for each node, in sequence.
    # For the moment, location objects are just tuples, in the order
    # "(x,y), i, s, alpha, (nx, ny)".
    def locations(self):
        # In the one-element list case, you can't divide by len-1.
        if len(self.data)==1:
            return [ (self.data[0], profile.Location(self.data[0].position(),
                                                     index=0, alpha=0.0) ) ]
        result = []
        for i in range(0, len(self.data)):
            node = self.data[i]
            result.append( (node,
                            profile.Location(node.position(),
                                             index=i,
                                             alpha=i*1.0/(len(self.data)-1) )
                                             ) )
        return result

