// -*- C++ -*-
// $RCSfile: equation.h,v $
// $Revision: 1.49 $
// $Author: langer $
// $Date: 2006-12-19 20:29:44 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// An Equation object is used to identify a set of rows in the global
// stiffness matrix and the rhs.  Each instance of the Equation class
// is a global variable: for example, HeatEqn, ForceBalanceEqn.  The
// instances are created in oof.py.

#include <oofconfig.h>

#ifndef EQUATION_H
#define EQUATION_H

class Equation;

#include "common/identification.h"
#include "common/pythonexportable.h"
#include "engine/fieldeqnlist.h"
#include "engine/indextypes.h"
#include "engine/sparselink.h"
#include "engine/linearizedsystem.h"
#include "common/space.h"
#include <string>
#include <vector>


class Edge;
class BoundaryEdge;
class EdgeGaussPoint;
class IndexP;
class IteratorP;
class Element;
class Field;
class FieldIndex;
class Flux;
class FluxData;
class FluxNormal;
class GaussPoint;
class CSubProblem;
class Material;
class NodalEquation;
class FuncNode;
class FluxNormal;
class ElementFuncNodeIterator;
class TimeDerivative;
int countEquations();
Equation *getEquationByIndex(int);


class Equation : public IdentifiedObject, public PythonExportable<Equation> {
private:
  int kinetic;
  int dynamic;
  /*
    kinetic and dynamic are set at object construction time.

    --If kinetic is true, the equation has a 
    \frac{\partial \phi}{\partial t} term.

    --If  dynamic is true, the equation has a 
    \frac{\partial^2 \phi}{\partial t^2} term.
    
    (\phi is an abstract (scalar ,vectorial, or tensorial) field)

    At the GUI level the appropriate button(s) will
    appear depending on their state.
    
    Application of the associated menu items enables the adequate
    construction of the C or M matrix.
   */
  bool kinetic_activity; //when true add first order time derivative(s) to Eqn
  bool dynamic_activity; //when true add second order time derivative(s) to Eqn
  std::vector<TimeDerivative*> timeDerivativeComponents; 
  // local list holding information of the pairs relating partial time
  // detivatives and the fields.
protected:
  int dim_;			// number of components
  std::string name_;
  static const std::string modulename_;
  Flux *fflux;  // Flux used by this equation.
  const unsigned int index_;
public:
  Equation(const std::string &nm, Flux &flx, int d);
  virtual ~Equation() {}
  unsigned int index() const { return index_; }
  static Equation *getEquation(const std::string&);

  virtual const std::string &classname() const = 0;
  virtual const std::string &modulename() const { return modulename_; }

  // where a given component lives in the eqn lists in a Node
  int localindex(const FuncNode&, const FieldIndex &component) const;
  int localindex(const FuncNode&, int component) const;
  NodalEquation *nodaleqn(const FuncNode&, int component) const;

  // Equations can return an iterator that iterates over their
  // components, which are the components of the divergence of the
  // host flux, for DivergenceEquations, and out-of-plane components
  // of the flux for PlaneFluxEquation objects.
  virtual IteratorP iterator() const = 0;
  virtual IndexP componenttype() const = 0;
  virtual IndexP getIndex(const std::string&) const = 0;

  static std::vector<Equation*> &all();
  const std::vector<TimeDerivative*> &allTimeDerivatives();
  const std::string &name() const { return name_; }

  const std::string &fluxname() const;
  const Flux* flux() const;

  int dim() const { return dim_; } // number of components
  int ndof() const { return dim_; }

  void activate_fluxes(CSubProblem*);	// called when Eqn is activated by CSubProblem
  void deactivate_fluxes(CSubProblem*);
  // This is synonymous with CSubProblem::is_active_equation(Equation&):
  bool is_active(const CSubProblem*) const;

  virtual void make_stiffness(CSubProblem *, const Element*, const GaussPoint&,
			      const Flux&, const FluxData*, 
			      LinearizedSystem&)
    const = 0;

  // Returns the polynomial order of the terms that the equation
  // factors into the stiffness matrix integrand.
  virtual int integration_order(const Element*) const = 0;

  // make_rhs is the body-force analogue of make_stiffness.  Not
  // immediately obvious if this should be pure-virtual....
  virtual void make_rhs(CSubProblem*, const Element*, const GaussPoint&,
			const Flux&, const FluxData*, LinearizedSystem&)
    const = 0;

  virtual void boundary_integral(const CSubProblem*, const Flux*, 
				 const BoundaryEdge*,
				 const EdgeGaussPoint &, const FluxNormal *)
    const = 0;

  // kinetic and dynamic are set by the programmer, and describe
  // whether the equation *can* be solved kinetically or dynamically.
  void set_kinetic(int k) {kinetic = k;}
  void set_dynamic(int d) {dynamic = d;}
  int get_kinetic() {return kinetic;}
  int get_dynamics() {return dynamic;}

  
  void addTimeDerivative(TimeDerivative *);

  virtual bool allow_boundary_conditions() const = 0;

  // Stuff required by fieldeqnlist.h
  typedef FieldEqnData FEData;	// just use base class from fieldeqnlist.h
  class FindAllEquationWrappers {
  private:
    FEMesh *mesh;
  public:
    FindAllEquationWrappers(FEMesh *mesh) : mesh(mesh) {}
    // Using typedefs to specify the return type of operator()
    // confuses the compiler, so we have to write it out in
    // full. Grrr.
    std::map<const std::vector<FEData>*, FEWrapper<Equation>*,
	     ltFEvector<Equation> >
    &operator()();
  };
  typedef FindAllEquationWrappers GetWrappers;
};


std::ostream &operator<<(std::ostream &os, const Equation&);

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// divergence of flux = body force
class DivergenceEquation : public Equation {
public:
  DivergenceEquation(const std::string &name, Flux &flx, int d); 
  virtual void make_stiffness(CSubProblem*, const Element*, const GaussPoint&,
			      const Flux&, const FluxData*, 
			      LinearizedSystem&)
    const;
  virtual void make_rhs(CSubProblem*, const Element*, const GaussPoint&,
			const Flux&, const FluxData*, LinearizedSystem&) const;
  virtual void boundary_integral(const CSubProblem*, const Flux*, 
				 const BoundaryEdge *, const EdgeGaussPoint&, 
				 const FluxNormal *) 
    const;
  virtual int integration_order(const Element*) const;

  virtual IteratorP iterator() const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;

  virtual const std::string &classname() const;
  virtual bool allow_boundary_conditions() const { return true; }
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class PlaneFluxEquation : public Equation {
public:
  PlaneFluxEquation(const std::string &name, Flux &flx, int d);
  virtual void make_stiffness(CSubProblem*, const Element*, const GaussPoint&,
			      const Flux&, const FluxData*, 
			      LinearizedSystem &)
    const;
  virtual void make_rhs(CSubProblem *, const Element *, const GaussPoint &,
			const Flux&, const FluxData*, LinearizedSystem&) const;
  virtual void boundary_integral(const CSubProblem*, const Flux*, 
				 const BoundaryEdge*, const EdgeGaussPoint&, 
				 const FluxNormal *)
    const;
  virtual int integration_order(const Element*) const;

  virtual IteratorP iterator() const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;

  virtual const std::string &classname() const;
  virtual bool allow_boundary_conditions() const { return false; }
};





#endif
