// -*- C++ -*-
// $RCSfile: steppingscheme.C,v $
// $Revision: 1.39 $
// $Author: langer $
// $Date: 2006-12-19 20:29:44 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "engine/csrmatrix.h"
#include "engine/csubproblem.h"
#include "engine/matvec.h"
#include "engine/solver.h"
#include "engine/solverdriver.h"
#include "engine/sparselink.h"
#include "engine/steppingscheme.h"

// TODO DEVEL: Encapsulate this better in the CSR.
#include "engine/SparseLib++/comprow_double.h"

void SteppingScheme::set_x(VECTOR_D *x) {
  if(x_copy == 0)
    x_copy = new VECTOR_D(x->size(), 0.0);
  else {
    if(x_copy->size() != x->size())
      x_copy->newsize(x->size());
  }
  *x_copy = *x;
}

SteppingScheme::~SteppingScheme() {
  if(x_copy)
    delete x_copy;
}


void QuasiStatic::update_mesh(CSubProblem &subproblem) {
  subproblem.update_time(get_dt());
}


void Kinetic::precompute(CSubProblem *subproblem,
			 CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x)
{
  //std::cerr << "entering  Kinetic::precompute" << std::endl;
  RWLock * rwlock = subproblem->get_rwlock();
  rwlock->read_acquire(); //acquire read lock
  try {
    //std::cerr << "after acquiring the lock" << std::endl;
    int n = A->nrows();
    //std::cerr << "after grabbing matrix size" << std::endl;
    if(conductivityMatrix) {
      delete conductivityMatrix;
      //std::cerr << "after deleting conductivityMatrix" << std::endl;
    }
    conductivityMatrix = new CSRmatrix(subproblem->ConductivityMatrix());
    if(!xn) {
      xn = new VECTOR_D(n, 0.0);
      //std::cerr << "after allocating xn" << std::endl;
    }
    //std::cerr << "x= " << *x << std::endl;
    (*xn) = (*x); //storing old solution
    if(!rhsn) {
      rhsn = new VECTOR_D(n, 0.0);
      //std::cerr << "a	fter allocating rhsn" << std::endl;
    }
    (*rhsn) = (*rhs); //storing old solution
  }
  catch(...) {
    rwlock->read_release(); 
    throw;
  }
  rwlock->read_release(); 
  //std::cerr << "leaving  Kinetic::precompute" << std::endl;
}

//get_K and get_b function set-up K x  = b
//with
//K = ConducivityMatrix + theta*dt*A
//and
//b = (ConducivityMatrix-(1.0-theta)*dt*A)*xn -dt*f_hat
//f_hat = rhs + theta*(rhs-rhsn)
//
//See Zienkiewicz volume 1, page 496, eq. 18.16 for more details
//
// Except that we are solving C.xdot + A.x = r, so our rhs has a 
// different sign.

// Also, the conductivity matrix only has in-plane components, so to
// (try to) reduce the stiffness of the resulting system, the thing
// TODO DEVEL is to instead construct a "hybrid" matrix.  Write A as
// [[alpha, beta], [gamma, delta]], where alpha is the all-in-plane
// part, delta is the all-out-of-plane part, and the other two are the
// obvious cross terms.  Then we may write the equation of motion as:
//
//   C x-dot + alpha.x + beta.y = r
//             gamma.x + delta.y = 0
//
// ...where x are the in-plane DOFs, and y the out-of-plane ones.  The
// RHS is always zero in the out-of-plane parts, because there can be
// no boundary conditions or applied forces there.  Iterating this
// forward in time in the Zienkiewicz scheme gets you a K matrix that
// looks like:
//
// K = [ C + alpha.dt.theta  beta.alpha.dt ]
//     [      gamma            delta       ]
//
//   This matrix is much less stiff, since the "delta" part is not
// multiplied by dt.  One can go ahead and solve this (with the
// appropriate rhs) to iterate the plane-stress system.
//
//   If that doesn't work, one can instead iterate x with fixed y, and
// then update y with the new, iterated x.
//
// In either case, the missing functionality is component addition of
// the matrices.  The reason for deferring all this stuff is to review
// the architecture of the matrix algebra, and decide what the Right
// Way(tm) to do things actually is.

// Addendum: This may not actually work very well in the general case
// -- for viscoelasticity, for example, the matrix which multiplies
// the time-derivatives of the DOFs has, in general, the same
// structure as its static counterpart, including the same planarity
// issues.

// TODO: Why is A passed-in here?  Couldn't the pointer to A be saved
// when it's passed in to precompute?  That way, conductivityMatrix
// and A would be treated equivalently.
// TODO: Why is subproblem passed in?
CSRmatrix * Kinetic::get_K(CSubProblem *subproblem, CSRmatrix *A)
{
  //std::cerr << "entering  Kinetic::get_K" << std::endl;

  // TODO DEVEL: Weird -- "dt" is class data -- this declaration assigns the
  // class variable's value to a "dt" which shadows the class
  // variable.
  double dt = get_dt();

//   CSRmatrix * thetadtA = new CSRmatrix((theta*dt)*(*A));
//   CSRmatrix * K = new CSRmatrix((*thetadtA) + (*conductivityMatrix));
//   delete thetadtA;
//   //std::cerr << "leaving  Kinetic::get_K" << std::endl;
//   return K;

  CSRmatrix *thetadtA;  
  if (theta == 0) {
    _K = new CSRmatrix(*conductivityMatrix);
    return _K;
  }
  if (theta == 1) {
    thetadtA = new CSRmatrix(dt*(*A));
    _K = new CSRmatrix((*thetadtA) + (*conductivityMatrix));
    delete thetadtA;
    return _K;
  }
  thetadtA = new CSRmatrix((theta*dt)*(*A));
  _K = new CSRmatrix((*thetadtA) + (*conductivityMatrix));
  delete thetadtA;
  return _K;
}


VECTOR_D * Kinetic::get_b(CSubProblem *subproblem,
			  CSRmatrix *A, VECTOR_D * rhs)
{
  //std::cerr << "entering  Kinetic::get_b" << std::endl;
  double dt = get_dt();
  CSRmatrix * oneminusthetadtA;
  VECTOR_D * f_hat;
  if (theta==0) {
    oneminusthetadtA = new CSRmatrix((dt)*(*A));
    // f_hat = new VECTOR_D((*rhsn) + (*rhs));
    // TODO: Why are f_hat and oneminusthetadtA allocated?  Why not just
    // use local variables?
    f_hat = new VECTOR_D( *rhsn );
    _b = new VECTOR_D((*conductivityMatrix)*(*xn) - (*oneminusthetadtA)*(*xn)
		      + dt*(*f_hat));
      delete f_hat;
      delete oneminusthetadtA;
      return _b;
  }
  if (theta == 1) {
    f_hat = new VECTOR_D( (*rhs) );
    _b = new VECTOR_D((*conductivityMatrix)*(*xn) + dt*(*f_hat));
    delete f_hat;
    return _b;
  }
  oneminusthetadtA = new CSRmatrix(((1.0-theta)*dt)*(*A));
  // f_hat = new VECTOR_D((1.0-theta)*(*rhsn) + (*rhs));
  f_hat = new VECTOR_D( (1.0-theta)*(*rhsn) + theta*(*rhs) );
  _b = new VECTOR_D((*conductivityMatrix)*(*xn) - (*oneminusthetadtA)*(*xn)
		    + dt*(*f_hat));
  delete f_hat;
  delete oneminusthetadtA;
  //std::cerr << "leaving  Kinetic::get_b" << std::endl;
  return _b;
}


void Kinetic::postcompute(CSubProblem *) {
  //std::cerr << "entering  Kinetic::postcompute" << std::endl;
  //std::cerr << "leaving  Kinetic::postcompute" << std::endl;
}


void Kinetic::clean_up() {
  if (_K) {
    delete _K;
    _K=0;
  }
  if (_b) {
    delete _b;
    _b=0;
  }
}

Kinetic::~Kinetic()
{
  //std::cerr << "entering Kinetic::~Kinetic()" << std::endl;
  delete xn;
  delete rhsn;
  delete conductivityMatrix;
  //std::cerr << "leaving Kinetic::~Kinetic()" << std::endl;
}


void Kinetic::update_mesh(CSubProblem &subproblem) {
  subproblem.update_time(get_dt());
}



//THIS SECTION IS FOR DEBUGGING
void sum_matrices() {
  int n_debug = 2;
  SparseLinkMatrix<double> A(n_debug,n_debug);
  A(0,0) = 1.0;
  A(0,1) = 1.0;
  A(1,1) = 1.0;
  //A(2,2) = 1.0;
  //  std::cerr << "Aslm=" << A << std::endl;
  SparseLinkMatrix<double> B(n_debug,n_debug);
  B(0,0) = 1.0;
  B(0,1) = 1.0;
  //B(1,0) = 1.0;
  //B(2,2) = 1.0;
 
  //  std::cerr << "Bslm=" << B << std::endl;
  SPARSE_MATRIX_D Asm(A);
  SPARSE_MATRIX_D Bsm(B);
  CSRmatrix Acsr (Asm);
  //std::cerr << "A=" << Acsr << std::endl;
  CSRmatrix Bcsr (Bsm);
  //std::cerr << "B=" << Bcsr << std::endl;
  CSRmatrix Ccsr (Acsr + Bcsr);
  //std::cerr << "sum=" << Ccsr << std::endl;

}

