// -*- C++ -*-
// $RCSfile: timederivative.C,v $
// $Revision: 1.17 $
// $Author: langer $
// $Date: 2006-12-19 20:29:44 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "engine/ooferror.h"
#include "common/switchboard.h"
#include "common/trace.h"
#include "common/tostring.h"
#include "engine/edge.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/field.h"
#include "engine/fieldindex.h" // Debugging.
#include "engine/flux.h"
#include "engine/fluxnormal.h"
#include "engine/gausspoint.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/property.h"
#include "engine/sparselink.h"
#include "engine/csubproblem.h"
#include "engine/timederivative.h"



TimeDerivative::TimeDerivative() : prop_category(0),
				   order(0),
				   equation(0),
				   fieldList(0)
{}

TimeDerivative::~TimeDerivative() {
}

void TimeDerivative::add_prop_category(const std::string &prop)
{
  prop_category.push_back(prop);
}


void TimeDerivative::add_field(const std::string &fld)
{
  fieldList.push_back(fld);
}

const std::vector<const Field*> TimeDerivative::get_fields() const {
  std::vector<const Field*> tdfields;
  for (std::vector<std::string>::size_type i = 0; i<fieldList.size(); i++)
    {
      tdfields.push_back(Field::getField(fieldList[i]));
    }
  return tdfields;
}

void TimeDerivative::make_matrix(CSubProblem *mesh, 
				 const Element *el,
				 const GaussPoint &gpt,
				 LinearizedSystem &system) const
{ 
  //if(mesh->kinetics_activated() && get_order() == 1)
  if (system.is_kinetic()) {
    //first order time derivative contributions
    make_conductivity_matrix(mesh, el, gpt, system);
  }
  //  if(mesh->dynamics_activated() && get_order() == 2) 
  
  // TODO DEVEL: This can't be right, why are we calling
  // make_conductivity_matrix again?  This should be a call to
  // make_mass_matrix or make_intertia_matrix or something.  Is this a
  // stub?
  if (system.is_dynamic()) {
    //second order time derivative contributions
    make_conductivity_matrix(mesh, el, gpt, system);
  }
}

const std::vector<Property*> TimeDerivative::get_properties(const Material &mat)
{
  //  std::cerr << " inside TimeDerivative::get_properties" << std::endl;
  std::vector<Property*> props;
  for (std::vector<std::string>::size_type jcat = 0; jcat<prop_category.size(); jcat++)
    {
      const std::string & prop_cat = prop_category[jcat];
      Property * matprop = mat.fetchProperty(prop_cat);
      props.push_back(matprop);
    }
  return props;
}

void TimeDerivative::make_conductivity_matrix(CSubProblem *mesh, 
					      const Element *el,
					      const GaussPoint &gpt,
					      LinearizedSystem &sys) const
{
  SparseLinkMatrix<double> &cmtx = sys.get_kinetic();

  // loop over nodes.  These are the test functions.
  for(ElementFuncNodeIterator node=el->funcnode_iterator();
      !node.end();++node) {
    // loop over components of time derivative
    for(int eqcomp=0; eqcomp<equation->dim(); eqcomp++) {
      /*
	eqcomp corresponds to an index that runs from 0 to the
	number of equation components in the Equation object,
	and NOT to the NodalEquation component in the node.
      */
      // row in global matrix 
      int rowno = mesh->map_forward_index(*equation, *node.funcnode(), eqcomp); //returns symmetrized row component
      
      // This row is an in-plane equation.
      sys.rowgroup[rowno] = -1;

      // loop over degrees of freedom of the element that couple to
      // this time derivative term by looping over all components of
      // the relevant *defined* fields at all the nodes.  The
      // out-of-plane fields are NOT included in the field list.
      const std::vector<const Field*> &fieldlist = get_fields();
      for(std::vector<const Field*>::size_type f=0; f<fieldlist.size(); f++) {
	const Field *field = fieldlist[f];
	if(mesh->is_defined_field(*field) && mesh->is_active_field(*field) &&
	   field->in_plane_part())
	  {
	    /*
	      READ THIS: Only in-plane fields contribute to the
	      conductivity/mass matrix, so in order to enforce this,
	      an additional conditional is added to the above
	      if-statement. When this formulation is extended to 3D,
	      such constraint will not be necessary.

	      TODO DEVEL: This is false -- in the general case,
	      e.g. viscoelasticity, the coefficient multiplying strain
	      rates is a full fourth-rank tensor object, just like a
	      Cijkl, with all the same planarity issues.
	    */
	  
	    for(ElementFuncNodeIterator nj=el->funcnode_iterator();!nj.end();
		++nj)
	      {
		// loop over components of field
		for(IteratorP fi=field->iterator(ALL_INDICES); !fi.end(); ++fi){
		  DegreeOfFreedom *dof = (*field)(nj, fi.integer());
		  // The integrals over the element area are
		  // accumulated directly in the conductivity/mass
		  // matrix.  The gauss point weight is included here.
		  double wgt = gpt.weight();
		  double raw = node.shapefunction(gpt)*nj.shapefunction(gpt);
		
		  // TODO DEVEL: What if property is space-dependent?  
		  double res = raw * wgt * fetch_property(*field,
							  *el->material());
		  cmtx(rowno, dof->dofindex()) += res;

		  // Mark this column as either in-plane or part of a
		  // group.  TODO: Note that out-of-plane components
		  // have already been filtered out in the conditional
		  // above, so the second part of this conditional is
		  // never executed.
		  sys.colgroup[dof->dofindex()] = 
		    (field->in_plane_part() ? -1 : nj.node()->index() );
		}
	      }
	  }
      }
    }
  }
}



ScalarKineticScalarTimeDerivative::ScalarKineticScalarTimeDerivative(const std::string &fld, const std::string & prop) : 
  TimeDerivative()
{
  add_prop_category(prop);
  add_field(fld);
  set_order(1);
}

ScalarKineticScalarTimeDerivative::~ScalarKineticScalarTimeDerivative() 
{

}

double ScalarKineticScalarTimeDerivative::fetch_property(const Field &f, const Material &mat) const 
{ 
  for (std::vector<std::string>::size_type item = 0; item<fieldList.size(); item++)
    {
      if (f.name()== fieldList[item])
	{
	  const std::string & prop_cat = prop_category[item];
	  ScalarTimeDerivPrefactor * matprop = dynamic_cast<ScalarTimeDerivPrefactor*>(mat.fetchProperty(prop_cat));
	  return matprop->fetch_value();
	}
    }
  throw ErrProgrammingError("Couldn't find property in local list of categories!!!",
			    __FILE__, __LINE__);
  return 1.0;//this line will never be reached but it is put in so the SGI compiler doesn't complain
}

ScalarDynamicVectorialTimeDerivative::ScalarDynamicVectorialTimeDerivative(const std::string &fld, const std::string & prop) : 
  TimeDerivative()
{
  add_prop_category(prop);
  add_field(fld);
  set_order(2);
}

ScalarDynamicVectorialTimeDerivative::~ScalarDynamicVectorialTimeDerivative() 
{
  
}

double ScalarDynamicVectorialTimeDerivative::fetch_property(const Field &f, const Material &mat) const
{
  for (std::vector<std::string>::size_type item = 0; item<fieldList.size(); item++)
    {
      if (f.name()== fieldList[item])
	{
	  const std::string & prop_cat = prop_category[item];
	  ScalarTimeDerivPrefactor * matprop = dynamic_cast<ScalarTimeDerivPrefactor*>(mat.fetchProperty(prop_cat));
	  return matprop->fetch_value();
	}
    }
  throw ErrProgrammingError("Couldn't find property in local list of categories!!!",
			    __FILE__, __LINE__);
  return 1.0;//this line will never be reached but it is put in  so the SGI compiler doesn't complain
}

